#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

use strict;

require Exporter;
require AutoLoader;

package W3C::Util::Exception;

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw(throw catch);
@EXPORT_OK = qw();
$VERSION = 0.91;

$W3C::Util::Exception::revision = '$Id: Exception.pm,v 1.10 1999/12/14 23:04:34 eric Exp $ ';

@W3C::Util::ChainedException::ISA = qw(W3C::Util::Exception);
@W3C::Util::IOException::ISA = qw(W3C::Util::Exception);
@W3C::Util::FileNotFoundException::ISA = qw(W3C::Util::IOException);
@W3C::Util::NotImplementedException::ISA = qw(W3C::Util::Exception);

#####
# per-class data
# hash table of current exceptions - needed in multi-threaded environment.
%W3C::Util::Exception::currentException = ();

#####
# constants
my ($PACKAGENAME, $FILENAME, $LINE, $SUBROUTINE, $HASARGS, $WANTARRAY, $EVALTEXT, $IS_REQUIRE) = (0..7);

#####
# throw

sub throw {
    my ($exception) = @_;
    $W3C::Util::Exception::currentExceptions{$exception} = $exception;
    die '<'.$exception.'> '.$exception->toString;
}

#####
# catch

sub catch {
    my ($type) = @_;
    return undef if ($@ !~ m/^\<([^\>]+)\>/);
    my $lookFor = $1;
    my $current = $W3C::Util::Exception::currentExceptions{$lookFor};
    return 0 if (!defined $current);
    if ($current->isa($type)) {
	delete $W3C::Util::Exception::currentExceptions{$lookFor};
	undef $@;
	return $current;
    } else {
	return undef;
    }
}

sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self;
    if (ref $parms eq 'HASH') {
	$self = $parms;
    } elsif (ref $parms eq 'ARRAY') {
	$self = {@$parms};
    } elsif (ref $parms eq 'SCALAR') {
	$self = {-message => $$parms};
    } elsif ($parms) {
	$self = {-message => $parms};
    } else {
	$self = {};
    }
    bless ($self, $class);
    $self->fillInStackTrace;
    return $self;
}

sub fillInStackTrace {
    my ($self) = @_;
    $self->{STACK_FRAMES} = [];
    for (my $i = 1; $i < 100; $i++) {
	my (@frame) = caller($i);
	if (defined $frame[$SUBROUTINE]) {
	    push (@{$self->{STACK_FRAMES}}, \@frame);
	} else {
	    last;
	}
    }
}

sub getMessage {return $_[0]->getSpecificMessage ? $_[0]->getSpecificMessage : $_[0]->{-message};}
sub toString {return $_[0]->getMessage;}
# overload this baby for exception-specific messages
sub getSpecificMessage {return undef;}
sub printStackTrace {
    my ($self, $outputHandle) = @_;
    my @messages;
    $self->printMessage($self->toString, $outputHandle);
    for (my $i = 0; $i < @{$self->{STACK_FRAMES}}; $i++) {
	my $frame = $self->{STACK_FRAMES}[$i];
	my @messageLine;
	if ($i > 0) {
	    if ($frame->[$SUBROUTINE] eq '(eval)' && $frame->[$EVALTEXT]) {
		push (@messageLine, 'eval ('.$frame->[$EVALTEXT].')');
	    } else {
		push (@messageLine, $frame->[$SUBROUTINE]);
	    }
	    push (@messageLine, '(...)') if ($frame->[$HASARGS]);
	} else {
	    push (@messageLine, (ref $self).' thrown');
	}
	push (@messageLine, ' at '.$frame->[$FILENAME].':'.$frame->[$LINE]);
	$self->printMessage(join ('', @messageLine));
    }
}

sub printMessage {
    my ($self, $message, $outputHandle) = @_;
    if (ref $outputHandle && $outputHandle->can('println')) {
	$outputHandle->println($message);
    } elsif (defined $outputHandle) {
	print $outputHandle $message."\n";
    } else {
	print STDERR $message."\n";
    }
}

# may make this cooler, but exceptions within exceptions are a pain
sub missingParm {
    my ($self, $name) = @_;
    return new W3C::Util::ChainedException({-missingParm => $name, -chainedException => $self});
}

package W3C::Util::ChainedException;
sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($parms);
    die (ref $self).' via (W3C::Util::ChainedException) exception needs a "-missingParm" parameter' 
	if (!$self->{-missingParm});
    die (ref $self).' via (W3C::Util::ChainedException) exception needs a "-chainedException" parameter' 
	if (!$self->{-chainedException});
    return $self;
}
sub getSpecificMessage {return (ref $_[0]->{-chainedException}).' exception needs a "'.$_[0]->{-missingParm}.'" parameter';}

package W3C::Util::FileNotFoundException;
sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($parms);
    $self->missingParm('-filename') if (!$self->{-filename});
    $self->{-error} = $! if (!$self->{-error});
    $self->fillInStackTrace;
    return $self;
}
sub getSpecificMessage {return "Can't open file \"".$_[0]->{-filename}.'": '.$_[0]->{-error};}
sub getFilename {return $_[0]->{-filename}}
sub getError {return $_[0]->{-error}}

package W3C::Util::NotImplementedException;
sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($parms);
    if (!$self->{-class} || !$self->{-method}) {
	my ($packageName, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require) = caller(1);
	$self->{-class} = $packageName if (!$self->{-class});
	$self->{-method} = $subroutine if (!$self->{-method});
    }
    $self->fillInStackTrace;
    return $self;
}
sub getSpecificMessage {return 'function "'.$_[0]->{-class}.'::'.$_[0]->{-method}.' is not implemented';}
sub getClass {return $_[0]->{-class}}
sub getMethod {return $_[0]->{-method}}
sub getFunction {return $_[0]->{-class}.'::'.$_[0]->{-filename}}

package W3C::Util::Exception;

1;

__END__

=head1 NAME

W3C::Util::Exception - experimental Exception throw and catch

=head1 SYNOPSIS

    use W3C::Util::Exception;
    @NegotiationException::ISA = qw(W3C::Util::Exception);
    eval {
	&iNeed('it yesterday');
    }; if ($@) {if (my $ex = &catch('NegotiationException')) {
	print "Try again with larger stick or carrot.\n";
    } elsif (my $ex = &catch('W3C::Util::Exception')) {
	warn 'got some random exception: '.$ex->toString."\n";
	$ex->printStackTrace;
	die "\n";
    } else {die $@;}}

    sub iNeed {
	throw(new NegotiationException('What about my needs?'));
    }

=head1 DESCRIPTION

This probably won't be needed when die passes a first class object.

This module is part of the W3C::Utils CPAN module.

Create an exception class and derive it from W3C::Util::Exception.
Write a function that throws it.
Call the function from inside an eval.
At the end of the eval, add
    "if ($@) {"
Try each exception that could be thrown with
    "if (my $ex = &catch('NegotiationException')) {"
At the end, have an "
    "} elsif (my $ex = &catch('W3C::Util::Exception')) {"
to make sure you get all possible exceptions. (Note - this is one area
where language support could make sure you've covered all the bases.)
At the end, you may want to add
    "} else {die $@;}}"
to make sure that you catch any non-exception calls to die.

Uncaught exceptions may be found by putting an eval around your
program's main. Exceptions that are thrown and are caught by a an eval
will display with a leading block containing the exception's address
and toString value. For instance:
    "<W3C::Util::FileNotFoundException=HASH(0x8313e1c)> Can't
     open file "asdf": No such file or directory at 
     ../../../W3C/Util/Exception.pm line 34."
This should be save as long there are no '>'s allowed in object::print.

=head1 NOTES

I considered using a syntax like
    try (sub {
	&iNeed('it yesterday');
    }, 'NegotiationException', sub {
	print "Try again with larger stick or carrot.\n";
    }, 'W3C::Util::Exception', sub {
	warn 'got some random exception: '.$ex->toString."\n";
	$ex->printStackTrace;
	die "\n";
    });
but figured it would just complicate the control without improving the
interface significantly. Anyone have any good ideas here?

=head1 AUTHOR

Eric Prud'hommeaux <eric@w3.org>

=head1 SEE ALSO

perl(1).

=cut
